@@ -5,7 +5,7 @@
 
 #ifndef MUON_FUNCTIONS_STRING_H
 #define MUON_FUNCTIONS_STRING_H
-#include "functions/common.h"
+#include "lang/func_lookup.h"
 
 enum format_cb_result {
 	format_cb_found,
@@ -16,9 +16,11 @@ enum format_cb_result {
 
 bool version_compare(struct workspace *wk, uint32_t err_node, const struct str *ver, obj cmp_arr, bool *res);
 
-typedef enum format_cb_result ((*string_format_cb)(struct workspace *wk, uint32_t node, void *ctx, const struct str *key, uint32_t *elem));
+typedef enum format_cb_result(
+	(*string_format_cb)(struct workspace *wk, uint32_t node, void *ctx, const struct str *key, uint32_t *elem));
 
 bool string_format(struct workspace *wk, uint32_t err_node, obj str, obj *res, void *ctx, string_format_cb cb);
 
-extern const struct func_impl_name impl_tbl_string[];
+extern const struct func_impl impl_tbl_string[];
+extern const struct func_impl impl_tbl_string_internal[];
 #endif
