@@ -12,7 +12,10 @@
 
 #ifdef _WIN32
 #ifndef S_IRUSR
-#define S_IRUSR 0
+#define S_IRUSR 0400
+#endif
+#ifndef S_IWUSR
+#define S_IWUSR 0200
 #endif
 #ifndef S_ISDIR
 #define S_ISDIR(m) (((m) & S_IFMT) == S_IFDIR)
@@ -26,9 +29,14 @@ extern int opterr, optind, optopt;
 #include <unistd.h>
 #endif
 
+#include "lang/object.h"
+
 int os_getopt(int argc, char *const argv[], const char *optstring);
 
 // Returns the number of jobs to spawn.  This number should be slightly larger
 // than the number of cpus.
 uint32_t os_parallel_job_count(void);
+
+void os_set_env(const struct str *k, const struct str *v);
+const char *os_get_env(const char *k);
 #endif
