@@ -5,26 +5,36 @@
 
 #include "compat.h"
 
+#include <stdlib.h>
+#include <string.h>
+
 #include "args.h"
+#include "backend/ninja.h"
+#include "backend/output.h"
 #include "buf_size.h"
-#include "datastructures/arr.h"
 #include "error.h"
+#include "lang/object_iterators.h"
 #include "lang/string.h"
 #include "lang/workspace.h"
 #include "log.h"
 #include "meson_opts.h"
+#include "platform/assert.h"
 #include "platform/path.h"
 #include "platform/run_cmd.h"
 #include "version.h"
 
-#include <stdlib.h>
-#include <string.h>
-
 struct translate_meson_opts_ctx {
 	obj prepend_args;
 	obj stray_args;
 	obj argv;
 	bool help;
+
+	// introspect specfic
+	bool introspect_force_object;
+
+	// compile specific
+	const char *compile_chdir;
+
 	const char *subcommand;
 };
 
@@ -259,11 +269,7 @@ translate_meson_opts_test(struct workspace *wk, char *argv[], uint32_t argc, str
 		{ "gdb", .ignore = true },
 		{ "gdb-path", true, .ignore = true },
 		{ "repeat", true, .ignore = true },
-		{
-			"wrapper",
-			true,
-			.ignore = true,
-		},
+		{ "wrapper", true, .ignore = true },
 		{ "no-suite", .ignore = true },
 		{ "no-stdsplit", .ignore = true },
 		{ "print-errorlogs", .ignore = true },
@@ -364,19 +370,11 @@ translate_meson_opts_setup(struct workspace *wk, char *argv[], uint32_t argc, st
 		{ "D", true, .handle_as = opt_setup_define },
 
 		{ "prefix", true, "prefix" },
-		{
-			"bindir",
-			true,
-			"bindir",
-		},
+		{ "bindir", true, "bindir" },
 		{ "datadir", true, "datadir" },
 		{ "includedir", true, "includedir" },
 		{ "infodir", true, "infodir" },
-		{
-			"libdir",
-			true,
-			"libdir",
-		},
+		{ "libdir", true, "libdir" },
 		{ "libexecdir", true, "libexecdir" },
 		{ "localedir", true, "localedir" },
 		{ "localstatedir", true, "localstatedir" },
@@ -391,37 +389,17 @@ translate_meson_opts_setup(struct workspace *wk, char *argv[], uint32_t argc, st
 		{ "build.pkg-config-path", true, "pkg_config_path" },
 		{ "buildtype", true, "buildtype" },
 		{ "cmake-prefix-path", true, "cmake_prefix_path" },
-		{
-			"cross-file",
-			true,
-			.ignore = true,
-		},
-		{
-			"debug",
-			false,
-			"debug",
-		},
+		{ "cross-file", true, .ignore = true },
+		{ "debug", false, "debug" },
 		{ "default-library", true, "default_library" },
 		{ "errorlogs", .ignore = true },
 		{ "fatal-meson-warnings", .ignore = true },
-		{
-			"force-fallback-for",
-			true,
-			"force_fallback_for",
-		},
+		{ "force-fallback-for", true, "force_fallback_for" },
 		{ "install-umask", true, "insall_umask" },
-		{
-			"layout",
-			true,
-			"layout",
-		},
+		{ "layout", true, "layout" },
 		{ "native-file", true, .ignore = true },
 		{ "optimization", true, "optimization" },
-		{
-			"pkg-config-path",
-			true,
-			"pkg_config_path",
-		},
+		{ "pkg-config-path", true, "pkg_config_path" },
 		{ "pkgconfig.relocatable", .ignore = true },
 		{ "prefer-static", false, "prefer_static" },
 		{ "python.install-env", true, .ignore = true },
@@ -471,20 +449,215 @@ translate_meson_opts_setup(struct workspace *wk, char *argv[], uint32_t argc, st
 	return true;
 }
 
+enum meson_opts_introspect {
+	opt_introspect_file = 1,
+	opt_introspect_force_object,
+};
+
+static bool
+translate_meson_opts_introspect_callback(struct workspace *wk,
+	const struct meson_option_spec *spec,
+	const char *val,
+	struct translate_meson_opts_ctx *ctx)
+{
+	switch ((enum meson_opts_introspect)(spec->handle_as)) {
+	case opt_introspect_file:
+		obj_array_push(wk, ctx->argv, make_str(wk, spec->name));
+		break;
+	case opt_introspect_force_object: ctx->introspect_force_object = true; return true;
+	default: UNREACHABLE;
+	}
+
+	return true;
+}
+
+static bool
+translate_meson_opts_introspect(struct workspace *wk, char *argv[], uint32_t argc, struct translate_meson_opts_ctx *ctx)
+{
+	struct meson_option_spec opts[] = {
+		{ "h", .help = true },
+		{ "help", .help = true },
+		{ "a", .ignore = true },
+		{ "all", .ignore = true },
+		{ "i", .ignore = true },
+		{ "indent", .ignore = true },
+		{ "backend", true, .ignore = true },
+		{ "f", .handle_as = opt_introspect_force_object },
+		{ "force-object-output", .handle_as = opt_introspect_force_object },
+		{ "ast", .handle_as = opt_introspect_file },
+		{ "benchmarks", .handle_as = opt_introspect_file },
+		{ "buildoptions", .handle_as = opt_introspect_file },
+		{ "buildsystem-files", .handle_as = opt_introspect_file },
+		{ "compilers", .handle_as = opt_introspect_file },
+		{ "dependencies", .handle_as = opt_introspect_file },
+		{ "scan-dependencies", .handle_as = opt_introspect_file },
+		{ "installed", .handle_as = opt_introspect_file },
+		{ "install-plan", .handle_as = opt_introspect_file },
+		{ "machines", .handle_as = opt_introspect_file },
+		{ "projectinfo", .handle_as = opt_introspect_file },
+		{ "targets", .handle_as = opt_introspect_file },
+		{ "tests", .handle_as = opt_introspect_file },
+	};
+
+	if (!translate_meson_opts_parser(
+		    wk, argv, argc, ctx, opts, ARRAY_LEN(opts), translate_meson_opts_introspect_callback)) {
+		return false;
+	}
+
+	const char *build_dir = 0;
+	if (get_obj_array(wk, ctx->stray_args)->len) {
+		obj build;
+		obj_array_index(wk, ctx->stray_args, 0, &build);
+		build_dir = get_cstr(wk, build);
+	}
+
+	uint32_t i = 0, intro_len = get_obj_array(wk, ctx->argv)->len;
+	bool make_object = intro_len > 1;
+	if (ctx->introspect_force_object) {
+		make_object = true;
+	}
+
+	SBUF(out);
+
+	if (make_object) {
+		sbuf_push(wk, &out, '{');
+	}
+
+	obj v;
+	obj_array_for(wk, ctx->argv, v) {
+			SBUF(path);
+
+			if (make_object) {
+				sbuf_pushf(wk, &out, "\"%s\":", get_cstr(wk, v));
+			}
+
+			if (build_dir) {
+				path_push(wk, &path, build_dir);
+			}
+			path_push(wk, &path, output_path.introspect_dir);
+			path_push(wk, &path, "intro-");
+			sbuf_pushf(wk, &path, "%s.json", get_cstr(wk, v));
+
+			struct source src;
+			if (!fs_read_entire_file(path.buf, &src)) {
+				LOG_E("failed to introspect %s", get_cstr(wk, v));
+				return false;
+			}
+
+			sbuf_pushn(wk, &out, src.src, src.len);
+
+			if (make_object && i < intro_len - 1) {
+				sbuf_push(wk, &out, ',');
+			}
+
+			++i;
+	}
+
+	if (make_object) {
+		sbuf_push(wk, &out, '}');
+	}
+
+	printf("%s\n", out.buf);
+
+	exit(0);
+}
+
+enum meson_opts_compile {
+	opt_compile_chdir = 1,
+	opt_compile_jobs,
+	opt_compile_verbose,
+	opt_compile_clean,
+	opt_compile_ninja_args,
+};
+
+static bool
+translate_meson_opts_compile_callback(struct workspace *wk,
+	const struct meson_option_spec *spec,
+	const char *val,
+	struct translate_meson_opts_ctx *ctx)
+{
+	switch ((enum meson_opts_compile)(spec->handle_as)) {
+	case opt_compile_chdir:
+		ctx->compile_chdir = val;
+		break;
+	case opt_compile_jobs:
+		obj_array_push(wk, ctx->argv, make_str(wk, "-j"));
+		obj_array_push(wk, ctx->argv, make_str(wk, val));
+		break;
+	case opt_compile_verbose:
+		obj_array_push(wk, ctx->argv, make_str(wk, "-v"));
+		break;
+	case opt_compile_clean:
+		obj_array_push(wk, ctx->argv, make_str(wk, "-t"));
+		obj_array_push(wk, ctx->argv, make_str(wk, "clean"));
+		break;
+	case opt_compile_ninja_args:
+		obj_array_extend(wk, ctx->argv, str_split(wk, &WKSTR(val), 0));
+		break;
+	default: UNREACHABLE;
+	}
+
+	return true;
+}
+
+static bool
+translate_meson_opts_compile(struct workspace *wk, char *argv[], uint32_t argc, struct translate_meson_opts_ctx *ctx)
+{
+	struct meson_option_spec opts[] = {
+		{ "h", .help = true },
+		{ "help", .help = true },
+		{ "clean", .handle_as = opt_compile_clean },
+		{ "C", true, .handle_as = opt_compile_chdir },
+		{ "j", true, .handle_as = opt_compile_jobs },
+		{ "jobs", true, .handle_as = opt_compile_jobs },
+		{ "l", true, .ignore = true },
+		{ "load-average", true, .ignore = true },
+		{ "v", .handle_as = opt_compile_verbose },
+		{ "verbose", .handle_as = opt_compile_verbose },
+		{ "ninja-args", true, .handle_as = opt_compile_ninja_args },
+		{ "vs-args", true, .ignore = true },
+		{ "xcode-args", true, .ignore = true },
+	};
+
+	if (!translate_meson_opts_parser(
+		    wk, argv, argc, ctx, opts, ARRAY_LEN(opts), translate_meson_opts_compile_callback)) {
+		return false;
+	}
+
+	if (ctx->help) {
+		return true;
+	}
+
+	obj v;
+	obj_array_for(wk, ctx->stray_args, v) {
+		const struct str *s = get_str(wk, v);
+		const char *sep;
+		if ((sep = strchr(s->s, ':'))) {
+			v = make_strn(wk, s->s, sep - s->s);
+		}
+
+		obj_array_push(wk, ctx->argv, v);
+	}
+
+	LO("ninja args: %o\n", ctx->argv);
+
+	if (!ninja_run(wk, ctx->argv, ctx->compile_chdir, 0)) {
+		exit(1);
+	}
+
+	exit(0);
+}
+
 static const struct {
 	const char *name;
 	translate_meson_opts_func translate_func;
 } meson_opts_subcommands[] = {
 	{ "setup", translate_meson_opts_setup },
 	{ "configure", translate_meson_opts_setup },
-	{
-		"install",
-		translate_meson_opts_install,
-	},
-	{
-		"test",
-		translate_meson_opts_test,
-	},
+	{ "install", translate_meson_opts_install },
+	{ "test", translate_meson_opts_test },
+	{ "introspect", translate_meson_opts_introspect },
+	{ "compile", translate_meson_opts_compile },
 };
 
 static translate_meson_opts_func
@@ -564,7 +737,7 @@ translate_meson_opts(struct workspace *wk,
 	ctx.argv = ctx.prepend_args;
 	obj_array_extend_nodup(wk, ctx.argv, ctx.stray_args);
 
-	obj_fprintf(wk, log_file(), "args: %o\n", ctx.argv);
+	obj_lprintf(wk, "args: %o\n", ctx.argv);
 
 	const char *argstr;
 	join_args_argstr(wk, &argstr, new_argc, ctx.argv);
